searchNodes=[{"ref":"AyeSQL.html","title":"AyeSQL","type":"module","doc":"Aye /ʌɪ/ exclamation (archaic dialect): said to express assent; yes. AyeSQL is a small Elixir library for using raw SQL. Overview Inspired on Clojure library Yesql, AyeSQL tries to find a middle ground between those two approaches by: Keeping the SQL in SQL files. Generating Elixir functions for every query. Having named parameters and query composability easily. Using the previous query, we would create a SQL file with the following contents: -- name: get_day_interval -- docs: Gets days interval. SELECT generate_series( current_date - :days::interval, -- Named parameter :days current_date - interval &#39;1 day&#39;, interval &#39;1 day&#39; )::date AS date; -- name: get_avg_clicks -- docs: Gets average click count. WITH computed_dates AS ( :get_day_interval ) SELECT dates.date AS day, count(clicks.id) AS count FROM computed_date AS dates LEFT JOIN clicks AS clicks ON date(clicks.inserted_at) = dates.date WHERE clicks.link_id = :link_id -- Named parameter :link_id GROUP BY dates.date ORDER BY dates.date; In Elixir we would load all the queries in this file by doing the following: defmodule Queries do use AyeSQL defqueries(&quot;queries.sql&quot;) # File name with relative path to SQL file. end And then we could execute the query as follows: params = %{ link_id: 42, days: %Postgrex.Interval{secs: 864000} # 10 days } {:ok, result} = Queries.get_avg_clicks(params, run?: true)"},{"ref":"AyeSQL.html#__using__/1","title":"AyeSQL.__using__/1","type":"macro","doc":"Uses AyeSQL for loading queries. The available options are: runner: module handling the query. Any other option will be passed to the runner."},{"ref":"AyeSQL.html#defqueries/1","title":"AyeSQL.defqueries/1","type":"macro","doc":"Macro to load queries from a file. Let&#39;s say we have the file my_queries.sql with the following contents: -- name: get_user -- docs: Gets user by username SELECT * FROM users WHERE username = :username; We can load our queries to Elixir using the macro defqueries/1 as follows: defmodule Queries do use AyeSQL, repo: MyRepo defqueries(&quot;my_queries.sql&quot;) end We can now do the following to get the SQL and the ordered arguments: iex(1)&gt; Queries.get_user!(%{username: &quot;some_user&quot;}) {&quot;SELECT * FROM user WHERE username = $1&quot;, [&quot;some_user&quot;]} If we would like to execute the previous query directly, the we could do the following: iex(1)&gt; Queries.get_user!(%{username: &quot;some_user&quot;}, run?: true) %Postgrex.Result{...} We can also run the query by composing it with the Queries.run/1 function generated in the module e.g: iex(1)&gt; %{username: &quot;some_user&quot;} |&gt; Queries.get_user!() |&gt; Queries.run!() %Postgrex.Result{...}"},{"ref":"AyeSQL.html#defqueries/3","title":"AyeSQL.defqueries/3","type":"macro","doc":"Macro to load queries from a file and create a module for them. Same as defqueries/1, but creates a module e.g: use AyeSQL, repo: MyRepo defqueries(Queries, &quot;my_queries.sql&quot;) This will generate the module Queries and it&#39;ll contain all the SQL statements included in &quot;sql/my_queries.sql&quot;."},{"ref":"AyeSQL.html#t:query/0","title":"AyeSQL.query/0","type":"type","doc":"AyeSQL query statement."},{"ref":"AyeSQL.Core.html","title":"AyeSQL.Core","type":"module","doc":"This module defines the core functionality for AyeSQL."},{"ref":"AyeSQL.Core.html#create_queries/1","title":"AyeSQL.Core.create_queries/1","type":"function","doc":"Creates several queries from the contents of a file."},{"ref":"AyeSQL.Core.html#run?/0","title":"AyeSQL.Core.run?/0","type":"function","doc":"Whether the queries should run by default or not. You can set this in the configuration as: use Mix.Config config :ayesql, run?: true"},{"ref":"AyeSQL.Core.html#t:arguments/0","title":"AyeSQL.Core.arguments/0","type":"type","doc":"SQL query arguments."},{"ref":"AyeSQL.Core.html#t:ast/0","title":"AyeSQL.Core.ast/0","type":"type","doc":"AyeSQL AST."},{"ref":"AyeSQL.Core.html#t:content/0","title":"AyeSQL.Core.content/0","type":"type","doc":"Contents of the query."},{"ref":"AyeSQL.Core.html#t:docs/0","title":"AyeSQL.Core.docs/0","type":"type","doc":"Documentation of the query function."},{"ref":"AyeSQL.Core.html#t:fun_def/0","title":"AyeSQL.Core.fun_def/0","type":"type","doc":"Definition of the query in the AST."},{"ref":"AyeSQL.Core.html#t:name/0","title":"AyeSQL.Core.name/0","type":"type","doc":"Name of the query function."},{"ref":"AyeSQL.Core.html#t:options/0","title":"AyeSQL.Core.options/0","type":"type","doc":"AyeSQL query options."},{"ref":"AyeSQL.Core.html#t:parameters/0","title":"AyeSQL.Core.parameters/0","type":"type","doc":"AyeSQL query parameters."},{"ref":"AyeSQL.Core.html#t:query/0","title":"AyeSQL.Core.query/0","type":"type","doc":"AyeSQL query."},{"ref":"AyeSQL.Core.html#t:statement/0","title":"AyeSQL.Core.statement/0","type":"type","doc":"SQL query statements."},{"ref":"AyeSQL.Runner.html","title":"AyeSQL.Runner","type":"behaviour","doc":"This module defines an AyeSQL.Runner."},{"ref":"AyeSQL.Runner.html#__using__/1","title":"AyeSQL.Runner.__using__/1","type":"macro","doc":"Uses the AyeSQL.Runner behaviour."},{"ref":"AyeSQL.Runner.html#c:run/3","title":"AyeSQL.Runner.run/3","type":"callback","doc":"Callback to initialize the runner."},{"ref":"AyeSQL.Runner.Ecto.html","title":"AyeSQL.Runner.Ecto","type":"module","doc":"This module defines Ecto default adapter. Can be used as follows: defmodule MyQueries do use AyeSQL, repo: MyRepo defqueries(&quot;query/my_queries.sql&quot;) end"},{"ref":"AyeSQL.Runner.Postgrex.html","title":"AyeSQL.Runner.Postgrex","type":"module","doc":"This module defines Postgrex default adapter. Can be used as follows: defmodule MyQueries do use AyeSQL, runner: AyeSQL.Runner.Postgrex defqueries(&quot;query/my_queries.sql&quot;) end And given a connection to the database, then it can be used with the query options: iex&gt; MyQueries.get_user([id: id], run?: true, conn: connection) {:ok, ...}"},{"ref":"readme.html","title":"AyeSQL","type":"extras","doc":"AyeSQL Aye /ʌɪ/ exclamation (archaic dialect): said to express assent; yes. AyeSQL is a small Elixir library for using raw SQL."},{"ref":"readme.html#why-raw-sql","title":"AyeSQL - Why raw SQL?","type":"extras","doc":"Writing and running raw SQL in Elixir is not pretty. Not only the lack of syntax highlighting is horrible, but also substituting parameters into the query string can be unmaintainable e.g: query = &quot;&quot;&quot; SELECT hostname, AVG(ram_usage) AS avg_ram FROM server WHERE hostname IN ($1, $2, $3) AND location = $4 GROUP BY hostname &quot;&quot;&quot; arguments = [&quot;server_0&quot;, &quot;server_1&quot;, &quot;server_2&quot;, &quot;Barcelona&quot;] Postgrex.query(conn, query, arguments) Adding more hostnames to the previous query is a nightmare, editing strings to add the correct index to the query. Thankfully, we have Ecto, which provides a great DSL for generating database queries at runtime. The same query in Ecto e.g: servers = [&quot;server_0&quot;, &quot;server_1&quot;, &quot;server_2&quot;] location = &quot;Barcelona&quot; from s in &quot;server&quot;, where: s.location == ^location and s.hostname in ^servers, select: %{hostname: s.hostname, avg_ram: avg(s.ram_usage)} Pretty straightforward and maintainable. So, why raw SQL?. Though Ecto is quite good with simple queries, complex queries often require the use of fragments, ruining the abstraction and making the code harder to read e.g: Let&#39;s say we have the following SQL query to retrieve the click count of a certain type of link every day of the last X days. In raw SQL this could be written as: WITH computed_dates AS ( SELECT dates::date AS date FROM generate_series( current_date - $1::interval, current_date - interval &#39;1 day&#39;, interval &#39;1 day&#39; ) AS dates ) SELECT dates.date AS day, count(clicks.id) AS count FROM computed_dates AS dates LEFT JOIN clicks AS clicks ON date(clicks.inserted_at) = dates.date WHERE clicks.link_id = $2 GROUP BY dates.date ORDER BY dates.date; Where $1 is the interval (%Postgrex.Interval{} struct) and $2 is some link ID. The query is easy to understand and easy to maintain. The same query in Ecto could be written as: dates = &quot;&quot;&quot; SELECT generate_series( current_date - ?::interval, current_date - interval &#39;1 day&#39;, interval &#39;1 day&#39; )::date AS d &quot;&quot;&quot; from( c in &quot;clicks&quot;, right_join: day in fragment(dates, ^days), on: day.d == fragment(&quot;date(?)&quot;, c.inserted_at), where: c.link_id = ^link_id group_by: day.d, order_by: day.d, select: %{ day: fragment(&quot;date(?)&quot;, day.d), count: count(c.id) } ) The previous code is hard to read and hard to maintain: Not only knowledge of SQL is required, but also knowledge of the intricacies of using Ecto fragments. Queries using fragments cannot use aliases defined in schemas, so the code becomes inconsistent."},{"ref":"readme.html#overview","title":"AyeSQL - Overview","type":"extras","doc":"Inspired on Clojure library Yesql, AyeSQL tries to find a middle ground between those two approaches by: Keeping the SQL in SQL files. Generating Elixir functions for every query. Having named parameters and query composability easily. Using the previous query, we would create a SQL file with the following contents: -- name: get_day_interval -- docs: Gets days interval. SELECT datetime::date AS date FROM generate_series( current_date - :days::interval, -- Named parameter :days current_date - interval &#39;1 day&#39;, interval &#39;1 day&#39; ); -- name: get_avg_clicks -- docs: Gets average click count. WITH computed_dates AS ( :get_day_interval ) SELECT dates.date AS day, count(clicks.id) AS count FROM computed_date AS dates LEFT JOIN clicks AS clicks ON date(clicks.inserted_at) = dates.date WHERE clicks.link_id = :link_id -- Named parameter :link_id GROUP BY dates.date ORDER BY dates.date; In Elixir we would load all the queries in this file by creating the following module: defmodule Queries do use AyeSQL, repo: MyRepo defqueries(&quot;queries.sql&quot;) # File name with relative path to SQL file. end or using the macro defqueries/3: import AyeSQL, only: [defqueries: 3] defqueries(Queries, &quot;queries.sql&quot;, repo: MyRepo) Both approaches will create a module called Queries with all the queries defined in &quot;queries.sql&quot;. And then we could execute the query as follows: iex(1)&gt; params = %{ iex(1)&gt; link_id: 42, iex(1)&gt; days: %Postgrex.Interval{secs: 864_000} # 10 days iex(1)&gt; } iex(2)&gt; Queries.get_avg_clicks(params, run?: true) {:ok, [ %{day: ..., count: ...}, ... ]}"},{"ref":"readme.html#syntax","title":"AyeSQL - Syntax","type":"extras","doc":"A SQL file can have as many queries as you want as long as they are named. Before the query, add a comment with the keyword name:. This name will be used for the functions&#39; names e.g -- name: get_servers -- docs: Generates the functions get_servers/1 and get_servers/2 SELECT hostname FROM server; And optionally they can have: Named parameters: Identified by a : followed by the name of the parameter e.g: -- These functions receive a map or a Keyword with the parameter :hostname. -- name: get_server SELECT * FROM server WHERE hostname = :hostname; SQL queries calls: Identified by a : followed by the name of the query in the same file e.g: -- name: get_locations SELECT id FROM location WHERE region = :region; -- This will compose :get_locations with get_servers_by_location. -- The function will receive a map or a Keyword with the parameter :region. -- name: get_servers_by_location SELECT * FROM servers WHERE location_id IN ( :get_locations ); Named optional parameters: Identified by a :_ followed by the name of the parameter e.g: -- name: get_servers SELECT * FROM server WHERE hostname = :hostname :_by_location -- name: by_location AND location = :location then we could compose the queries as follow: For: SELECT * FROM server WHERE hostname = :hostname it would be called like this: Queries.get_servers(hostname: &quot;server0&quot;) For: SELECT * FROM server WHERE hostname = :hostname AND location = :location it would be called like this: Queries.get_servers( hostname: &quot;server0&quot;, location: &quot;Barcelona&quot;, :_by_location: &amp;Queries.by_location/2 ) Documentation: Before the query, add a comment with the keyword docs:. This string will be used as documentation for the function e.g: -- name: get_servers -- docs: Gets all the servers hostnames. SELECT hostname FROM server;"},{"ref":"readme.html#in-statement","title":"AyeSQL - IN statement","type":"extras","doc":"Let&#39;s say we have the following query loaded in the module Server: -- name: get_avg_ram SELECT hostname, AVG(ram_usage) AS avg_ram FROM server WHERE hostname IN (:hostnames) AND location = :location GROUP BY hostname; It is possible to do the following: iex(1)&gt; hosts = [&quot;server_0&quot;, &quot;server_1&quot;, &quot;server_2&quot;] iex(2)&gt; params = %{hostnames: {:in, hosts}, location: &quot;Barcelona&quot;} iex(3)&gt; Server.get_avg_ram(params, run?: true) {:ok, [ %{hostname: &quot;server_0&quot;, avg_ram: ...}, %{hostname: &quot;server_1&quot;, avg_ram: ...}, %{hostname: &quot;server_2&quot;, avg_ram: ...}, ]}"},{"ref":"readme.html#query-composability-at-runtime","title":"AyeSQL - Query composability at runtime","type":"extras","doc":"Let&#39;s say we have the following query loaded in the module Server: -- name: get_servers SELECT hostname FROM server WHERE region = :region; It is possible to do the following: iex(1)&gt; query = &amp;Server.get_servers/2 iex(2)&gt; params = %{hostnames: query, location: &quot;Barcelona&quot;, region: &quot;Spain&quot;} iex(3)&gt; Server.get_avg_ram(params, run?: true) {:ok, [ %{hostname: ...}, ... ]}"},{"ref":"readme.html#query-runners","title":"AyeSQL - Query runners","type":"extras","doc":"The purpose of runners is to be able to implement other database adapters. By default, AyeSQL uses the runner AyeSQL.Runner.Ecto. This runner only has one option which is :repo for the repo module. Additionally, it converts the result to a list of maps. Using other runners is as easy as setting them in the module definition as follows: defmodule MyQueries do use AyeSQL, runner: IdemRunner, repository: MyRepo ... end For runners, there is only one callback to be implemented. run/3: which receives a AyeSQL.statement(), AyeSQL.arguments() and a keyword() list with extra options for the runner (mandatory). The following would be a runner for Ecto that does nothing to the result (returns Postgrex.Result.t() and Postgrex.Error.t() structs): defmodule IdemRunner do use AyeSQL.Runner @impl true def run(stmt, args, options) do repo = options[:repository] || raise ArgumentError, &quot;No repo defined&quot; Ecto.Adapters.SQL.query(repo, stmt, args) end end"},{"ref":"readme.html#installation","title":"AyeSQL - Installation","type":"extras","doc":"AyeSQL is available as a Hex package. To install, add it to your dependencies in your mix.exs file: def deps do [{:ayesql, &quot;~&gt; 0.4&quot;}] end If you&#39;re going to use any of the provided query runners, then you should add their dependencies as well: Add :ecto_sql for AyeSQL.Runner.Ecto (default runner). Add :postgrex for AyeSQL.Runner.Postgrex. Add :ecto_sql and :postgrex for running queries using Ecto in a PostgreSQL database."},{"ref":"readme.html#author","title":"AyeSQL - Author","type":"extras","doc":"Alexander de Sousa."},{"ref":"readme.html#license","title":"AyeSQL - License","type":"extras","doc":"AyeSQL is released under the MIT License. See the LICENSE file for further details."}]